#!/usr/bin/perl -w
use strict;

# Equitable coloring checker.
#
# (c) 2008 Daniel E. Severin.

print "checkres - Equitable coloring checker\n\n";

sub manual {
	print "Usage: ./checkres.pl file.col file.res\n";
	exit 1;
}

if ($#ARGV < 1) { manual; }

my $arch_col = $ARGV[0];
my $arch_res = $ARGV[1];

open IN1, '<', $arch_col or die "Error reading $!";
open IN2, '<', $arch_res or die "Error reading $!";

$_ = <IN2>;
chomp;
my @datas = split ' ', $_;
if ($datas[0] ne 'CLRS') {
	print "Invalid RES file!\n";
	exit 1;
}

my $colors = $datas[1];
print "Colors used: $colors\n";

my $i;
my @col;
my $vertices = 0;
while (<IN2>) {
	chomp;
	@datas = split ' ', $_;
	for ($i = 0; $i <= $#datas; $i++) {
		my $c = $datas[$i];
		if ($c > $colors) {
			print "Invalid color found!\n";
			exit 1;
		}
		$vertices++;
		$col[$vertices] = $c;
	}
}

print "Verifying if the coloring is PROPER ..... ";
while (<IN1>) {
	chomp;
	@datas = split ' ', $_;
	if (defined $#datas) {
		if ($#datas >= 2) {
			if ($datas[0] eq 'e') {
				my $v1 = $datas[1];
				my $v2 = $datas[2];
				my $c = $col[$v2];
				if ($col[$v1] == $c) {
					print "Oops! Edge ($v1,$v2) has vertices with color $c  :(\n";
					exit 1;
				}
			}
		}
	}
}
print "[ OK ]\n";

print "Verifying if the coloring is EQUITABLE ..... ";
my @count;
for ($i = 1; $i <= $colors; $i++) { $count[$i] = 0; }
for ($i = 1; $i <= $vertices; $i++) { $count[$col[$i]]++; }
my $lowlimit = 1000000000;
my $lowcolor = 0;
my $uplimit = -1;
my $upcolor = 0;
for ($i = 1; $i <= $colors; $i++) {
	my $x = $count[$i];
	if ($x < $lowlimit) {
		$lowlimit = $x;
		$lowcolor = $i;
	}
	if ($x > $uplimit) {
		$uplimit = $x;
		$upcolor = $i;
	}
}
if ($lowlimit == $uplimit) {
	print "[ PERFECT ]\n";
	print "Every color is used $lowlimit times\n";
} else {
	if ($lowlimit == $uplimit - 1) {
		print "[ OK ]\n";
		print "Class colors have between $lowlimit and $uplimit vertices.\n";
		print "For instance, color $lowcolor is used $lowlimit times.\n";
		print "And color $upcolor is used $uplimit times.\n";
	}
	else {
		print "Oops! It is not equitable  :(\n";
		print "For instance, the color less used is $lowcolor and it is used $lowlimit times.\n";
		print "And the color more used is $upcolor and it is used $uplimit times.\n";
		exit 1;
	}
}

close IN1;
close IN2;
exit 0;
