
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * ecopt.h - main header
 ***/

#ifndef __ECOPT_H
#define __ECOPT_H

#ifdef __cplusplus
extern "C" {
#endif

/* Inclusions */

#include <cplex.h>
#ifdef ECOPARALLEL
#define __USE_GNU
#include <pthread.h>
#endif

/* Error codes */

#define ECOERR_CPLEX 0x800000
#define ECOERR_OK 0
#define ECOERR_MANYEDGES 1
#define ECOERR_MEMORY 2
#define ECOERR_NOENV 3
#define ECOERR_NOGRP 4
#define ECOERR_NOTVERTEX 5
#define ECOERR_LOOPS 6
#define ECOERR_OUTOFRANGE 7
#define ECOERR_NOTIMPLEMENTED 8
#define ECOERR_CREAT 9
#define ECOERR_INTERNAL 10
#define ECOERR_NULLP 11
#define ECOERR_BOUNDS 12
#define ECOERR_INCOMPAT 13
#define ECOERR_LASTCODE 13

#define ECOTABU_OK 10000
#define ECOTABU_TENURETOOHIGH 10001
#define ECOTABU_TIMEEXHAUSTED 10002
#define ECOTABU_ITERLIMIT 10003

#define ECOENUM_OK 20000
#define ECOENUM_INFEASIBLE 20001
#define ECOENUM_TIMEEXHAUSTED 20002
#define ECOENUM_OPTIMAL 20003

/* Options and parameters: general */

#define ECO_MAXTIME 7200
#define ECO_MAXFILENAMESIZE 128
#define ECO_MINVERTICES 5
#define ECO_MAXVERTICES 5000
#define ECO_MAXEDGES 2000000
#define ECO_EPSILON 0.0000001
#define ECO_WJTOL 0.0001
#define ECO_ISDYNAMICVAL 2 /* 0 = STATIC, 1 = DYN1, 2 = DYN2 */
#define ECO_MAXADYACSUST 3000
#define ECO_ENUMTIMELIMIT 999999.0
#define ECO_ENUMFACTOR 800

/* Options and parameters: cutting plane algorithm */

#define ECO_MAXFAMILYCUTROUND 200
#define ECO_MAXROUNDSATROOT 20
#define ECO_MAXROUNDSATNODE 1
#define ECO_VIOLATION 0.1
#define ECO_SKIPFACTORCUT 2
#define ECO_MAXDEPTHTREE 50

/* Options and parameters: heuristics */

#define ECO_TABUALPHA 90
#define ECO_TABUBETA 900
#define ECO_MAXTIMEINITABU 30.0
#define ECO_MAXTIMEPRITABU 5.0
#define ECO_MAXITER 50000
#define ECO_MAXITERPRI 2000
#define ECO_SKIPFACTORHEUR 1
#define ECO_MAXTIMEHEUR 300
#define ECO_THRESHOLD_VNUM 10

/* Options and parameters: specific */

#define ECO_OFF 0
#define ECO_ON 1
#define ECO_ASCENDING 1
#define ECO_DESCENDING 2
#define ECO_RANDOMIZE 3

#define ECO_FILE_NONE 0
#define ECO_FILE_LPFILE 1
#define ECO_FILE_LTXFILE 2

#define ECO_UPPERB_DELTA 0
#define ECO_UPPERB_GREEDY 1
#define ECO_UPPERB_TABU 2
#define ECO_UPPERB_PTABU 3

#define ECO_STRATEGY_OFF 0
#define ECO_STRATEGY_BYB 1
#define ECO_STRATEGY_CPLEX 2
#define ECO_STRATEGY_BYC 3
#define ECO_STRATEGY_ENUM 4
#define ECO_STRATEGY_ENUMDYN1 5
#define ECO_STRATEGY_ENUMDYN2 6
#define ECO_STRATEGY_CPLEXUB 7

#define ECO_FORM_STANDARD 0
#define ECO_FORM_SYM1 1
#define ECO_FORM_SYM2A 2
#define ECO_FORM_SYM2B 3
#define ECO_FORM_SYM2C 4

#define ECO_BRANCH_CPLEXDEF 0
#define ECO_BRANCH_PSEUDOREDUCED 1
#define ECO_BRANCH_CALLBACK 2

#define ECO_NODESEL_DFS 0
#define ECO_NODESEL_BESTBOUND 1

#define ECO_CUTPLANE_BIT_ON 1
#define ECO_CUTPLANE_BIT_CLIQUE 2
#define ECO_CUTPLANE_BIT_ALPHA2 4
#define ECO_CUTPLANE_BIT_BLOCK 8
#define ECO_CUTPLANE_BIT_NEIGHBOR 16
#define ECO_CUTPLANE_BIT_OUTSIDENEIGHBOR 32
#define ECO_CUTPLANE_BIT_CS 64
#define ECO_CUTPLANE_BIT_UJKS 128
#define ECO_CUTPLANE_BIT_EXTRA 256
#define ECO_CUTPLANE_BIT_CPLEX 512

#define ECO_PRIHEUR_BIT_ON 1
#define ECO_PRIHEUR_BIT_SCANXVJ 2
#define ECO_PRIHEUR_BIT_SCANCOLOR 4
#define ECO_PRIHEUR_BIT_ONLYROOT 8
#define ECO_PRIHEUR_BIT_CPLEX 16
#define ECO_PRIHEUR_BIT_TABU 32

#define ECO_PARAM_TILIM 0
#define ECO_PARAM_ITLIM 1
#define ECO_PARAM_LOWERB 2
#define ECO_PARAM_STRATEGY 3
#define ECO_PARAM_UPPERB 4
#define ECO_PARAM_FORM 5
#define ECO_PARAM_ALPHA 6
#define ECO_PARAM_BETA 7
#define ECO_PARAM_USERUB 8
#define ECO_PARAM_BRANCH 9
#define ECO_PARAM_NODESEL 10
#define ECO_PARAM_MIPINTERVAL 11
#define ECO_PARAM_THREADS 12
#define ECO_PARAM_USERLB 13
#define ECO_PARAM_CUTPLANE 14
#define ECO_PARAM_PRIHEUR 15
#define ECO_PARAM_WRITEFILE 16
#define ECO_PARAM_SKIPFACTORCUT 17
#define ECO_PARAM_MAXROUNDSATROOT 18
#define ECO_PARAM_MAXROUNDSATNODE 19
#define ECO_PARAM_MAXDEPTHTREE 20
#define ECO_PARAM_MAXTIMEHEUR 21
#define ECO_PARAM_IMPENUM 22
#define ECO_PARAM_IMPVERT 23
#define ECO_PARAM_LASTCODE 23

/* Definitions and structures */

#define ECODEF_PRIMAL_NOCHECKFEAS
#define ECODEF_FIXCLIQUE
/* #define ECODEF_DONOTORDER */
#define ECODEF_PURGEABLE
/* #define ECODEF_EXITAFTERROOT */
/* #define ECODEF_ABRULE */
/* #define ECODEF_REPORTDEPTH */
/* #define ECODEF_CUTANDBRANCHINLOWDENSITY */

struct ecovec {
  int size, maxsize;
  int *elem;
};
typedef struct ecovec *ECOVECptr;
struct ecomtx {
  int size1, size2;
  int *elem;
};
typedef struct ecomtx *ECOMTXptr;
struct ecolist {
  int value;
  struct ecolist *next;
};
struct ecoenv {
  CPXENVptr cplexenv;
  int par[ECO_PARAM_LASTCODE+1];
  long long iter_result;
  long long nodes_result;
  int timeopt_result;
  int timecut_result;
  double bestobjval;
  CPXCHANNELptr cpxresults, cpxwarning, cpxerror, cpxlog;
  void(*msgfunction)(const char *);
  char *filename;
};
typedef struct ecoenv *ECOENVptr;
struct ecogrp {
  int vertices;
  int edges;
  int maxedges;
  ECOVECptr edge1;
  ECOVECptr edge2;
  ECOVECptr coloring;
  int colors_used;
  int is_optimal;
  struct ecolist *hash_vertices;
};
typedef struct ecogrp *ECOGRPptr;

/* Structures for internal use only! */

struct ecocut {
  struct ecointernal *ei;
  CPXCENVptr cenv;
  void *cbdata;
  int wherefrom;
  int *useraction_p;
  int *cutind;
  double *cutval;
  double rhs;
  char sense;
  int nzcnt;
};
struct ecointernal {
#ifdef ECOPARALLEL
  int threads; /* number of threads */
  pthread_t self; /* ID of the thread */
  int tstate; /* 0 = not defined, 1 = defined */
#endif
#ifdef ECODEF_EXITAFTERROOT
  int firstobjvalroot, bestobjvalroot, bestroundroot; /* used for benchmarking cutting-plane algorithm */
  double firsttimeroot, besttimeroot;  
#endif
  ECOENVptr env; /* for accessing environment */
  ECOGRPptr grp; /* for accessing graph */
  ECOVECptr *degree; /* degree and neighborhood of each vertex */
  ECOVECptr ascending_by_degree; /* ordering of vertices by degree: 0, 1, ... */
  ECOVECptr descending_by_degree; /* ordering of vertices by degree: delta, delta-1, ... */
  ECOVECptr sl_ordering; /* ordering of vertices by SL method */
  int count_isolated; /* number of isolated vertices */
  ECOVECptr neighbormaxindep1_size; /* lower bound of alpha(N(v)) */
  ECOVECptr neighbormaxindep2_size; /* upper bound of alpha(N(v)) */
  ECOVECptr clique_set; /* initial maximal clique */
  int density; /* density of graph (from 1 to 99) */
  int min_color; /* lower bound of chieq */
  int max_color; /* upper bound of chieq */
  int bestnumberofcolors; /* number of colors used by bestcoloring in ECOenumbranch */
  int curnumberofcolors; /* number of colors used by curcoloring in ECOenumbranch and ECOcheckineq */
  long long nodeprocessed, totalnodeprocessed; /* number of nodes processed in ECOenumbranch */
  int enumactivated; /* 1 = run implicit enumeration */
  double enuminittime, enumavailabletime; /* time before starting an enumeration and time available for remaining enumerations */
  double start_t, maxtim_t, last_t; /* time: "before starting", "maximum" and "last before logging" */
  int isdynamic; /* 0 = static, 1 = dynamic (color order:1,2,...,k+1), 2 = dynamic (color order:k+1,1,...,k) */
  int numberofsolutions; /* number of solutions of the polyhedron enumerated used by ECOcheckineq */
  int numberofsolutionsface; /* number of solutions of the face enumerated used by ECOcheckineq */
  int *array_v, *array_j, *array_coef, array_size, rhs; /* inequality used by ECOcheckineq */
  ECOVECptr ordering_symmetry; /* ordering of vertices according to symmetry.c */
  ECOMTXptr xvar_state; /* state of x_vj variables: 0 - fix0, 1 - fix1, 2 - var0, 3 - var1, ... */
  ECOVECptr wvar_state; /* state of w_j variables: idem. */
  CPXLPptr lp; /* IP formulation */
  int variables; /* number of variables */
  int constraints; /* number of constraints */
  int adyac_substitutes; /* 1 = adjacency constraints substituted by neighborhood constraints */
  FILE *rootstream; /* file for saving lowerbound progress at root node */
  ECOVECptr tabu_bestcoloring; /* used by tabu heuristics */
  ECOVECptr tabu_curcoloring; /* used by tabu heuristics */
  ECOVECptr tabu_conflictvertexset; /* used by tabu heuristics */
  ECOVECptr tabu_conflictvertexlist; /* used by tabu heuristics */
  ECOVECptr tabu_bigclasses; /* used by tabu heuristics */
  ECOVECptr tabu_coloredvert; /* used by tabu heuristics */
  ECOMTXptr tabu_tenure; /* used by tabu heuristics */
  int callbacks_made; /* number of callbacks */
  double branchcallback_time; /* CPU time consumed by branch callback (including enumeration) */
  double cutcallback_time; /* CPU time consumed by cut callback */
  double heurcallback_time; /* CPU time consumed by heur/incumbent callback */
  double *x; /* fractional value of (x, w) */
  double LB_current; /* current lower bound of chieq */
  double UB_current; /* current upper bound of chieq */
  double objective; /* value of sum w*_j */
  int is_wj_integer; /* 1 = every wj is integer */
  int first_j_frac, last_j_frac; /* first and last value such that 0 < wj < 1;
               if is_wj_integer = 1, then w_{last_j_frac} = 1, w_{first_j_frac} = 0 */ 
  int cut_enabled; /* 1 = cuts are enabled, 0 = cuts are disabled */
  int cut_enabled_node; /* 1 = cuts enabled at current node, 0 = disabled */
  int round; /* current rount (at root node) */
  int isroot; /* 1 = at root node, 0 = subproblem */
  int cuts_applied; /* number of cuts applied */
  int cuts_applied_round; /* number of cuts applied in the current round */
  struct ecocut cu; /* current cut */
  ECOVECptr cut_temp1; /* internal vector for sep. routine */
  ECOVECptr cut_temp2; /* internal vector for sep. routine */
  ECOVECptr cut_descending; /* vertices ordered by x*_vjs */
  ECOVECptr cut_forbidden; /* forbidden vertices when finding cliques/2-rank */
  int clique_cuts_applied; /* number of clique cuts applied */
  int clique_cuts_applied_round; /* number of clique cuts applied in the current round */
  ECOVECptr clique_cut_set; /* maximal clique found for the cut */
  ECOVECptr clique_ujks; /* used during ujks separation */
  int alpha2_cuts_applied; /* number of 2-rank cuts applied */
  int alpha2_cuts_applied_round; /* number of 2-rank cuts applied in the current round */
  ECOVECptr alpha2_cut_set; /* maximal set found for the cut (with coef 1) */
  ECOVECptr alpha2clique_cut_set; /* maximal set found for the cut (with coef 2) */
  ECOVECptr alpha2_cut_temp1; /* internal vector for sep. routine */
  ECOVECptr alpha2_cut_temp2; /* internal vector for sep. routine */
  ECOVECptr alpha2_cut_temp3; /* internal vector for sep. routine */
  int block_cuts_applied; /* number of block cuts applied */
  int block_cuts_applied_round; /* number of block cuts applied in the current round */
  int extra_cuts_applied; /* number of extra-symmetry cuts applied */
  int extra_cuts_applied_round; /* number of extra-symmetry cuts applied in the current round */
  int neighborhood_cuts_applied; /* number of neighborhood cuts applied */
  int neighborhood_cuts_applied_round; /* number of neighborhood cuts applied in the current round */
  int outsideneighbor_cuts_applied; /* number of outside-neighborhood cuts applied */
  int outsideneighbor_cuts_applied_round; /* number of outside-neighborhood cuts applied in the current round */
  int ujks_cuts_applied; /* number of ujks-clique-neighborhood cuts applied */
  int ujks_cuts_applied_round; /* number of ujks-clique-neighborhood cuts applied in the current round */
  int cs_cuts_applied; /* number of S-color cuts applied */
  int cs_cuts_applied_round; /* number of S-color cuts applied in the current round */
  ECOVECptr cs_cut_index; /* 'index' for sorting x*_vjs */
  ECOVECptr cs_cut_value; /* 'value' for sorting x*_vjs */
  ECOVECptr cs_cut_descending; /* vertices ordered by x*_vjs */
  ECOVECptr cs_cut_c_set; /* C set found for the cut */
  ECOVECptr cs_cut_s_set; /* S set found for the cut */
  ECOVECptr cs_cut_forbidden; /* forbidden vertices when finding CS cuts */
  int cut_executed; /* number of times cut callback was executed */
  int primal_enabled; /* 1 = primal heuristic is enabled, 0 = primal heuristic is disabled */
  int primal_enabled_node; /* 1 = primal heuristic is enabled at current node, 0 = disabled */
  ECOVECptr primal_coloring; /* coloring generated by primal heuristic and enumerative routine */
  ECOVECptr primal_vector1; /* used by ECOmakefeasible and ECOprigreedy */
  ECOVECptr primal_vector2; /* used by ECOmakefeasible and ECOprigreedy */
  ECOVECptr primal_vector3; /* used by ECOmakefeasible and ECOprigreedy */
  ECOVECptr primal_vector4; /* used by ECOmakefeasible and ECOprigreedy */
  ECOVECptr primal_vector5; /* used by ECOprigreedy */
  ECOVECptr primal_vector6; /* used by ECOprigreedy */
  int primal_colors_used; /* number of colors used in primal coloring */
  int enum_found_sol; /* 1 = if enumerative routine has found a solution */
  int primal_executed; /* number of times the heuristic was executed */
  int make_tabu; /* 1 = execute tabu search right now */
  int totaliter; /* iterations of tabu search during MIP optimization */
  int *intfeas; /* feasibility info of variables (used by branch.c) */
  ECOVECptr colority; /* colority of neighborhood (used by branch.c) */
  int *varind; /* variable indices (used by branch.c) */
  char *varlu; /* variable bound selection (used by branch.c) */
  int *varbd; /* variables bound value (used by branch.c) */
};

/* Function prototypes */

double ECOclock();
ECOENVptr ECOopen(int *status, void(*msgfunction)(const char *), char *filename);
int ECOclose(ECOENVptr *env);
void ECOgeterrorstring(ECOENVptr env, int errcode, char *errmsg);
void ECOabort(int errcode);
int ECOgetparam(ECOENVptr env, int param);
void ECOsetparam(ECOENVptr env, int param, int value);
ECOGRPptr ECOcreategraph(ECOENVptr env, int *status, int vertices, int edges);
int ECOfreegraph(ECOENVptr env, ECOGRPptr *grp);
int ECOaddedge(ECOENVptr env, ECOGRPptr grp, int u, int v);
int ECOgetedges(ECOENVptr env, ECOGRPptr grp);
int ECOshowgraph(ECOGRPptr grp);
int ECOwriteformulation(ECOENVptr env, ECOGRPptr grp);
int ECOoptimize(ECOENVptr env, ECOGRPptr grp);
int* ECOcolresult(ECOGRPptr grp, int *status, int *is_optimal, int *colors_used);
double ECOgetbestobjval(ECOENVptr env);
long long ECOgetitcnt(ECOENVptr env);
long long ECOgetnodecnt(ECOENVptr env);
int ECOgettimeopt(ECOENVptr env);
int ECOgettimecut(ECOENVptr env);
int ECOcheckineq(ECOENVptr env, ECOGRPptr grp, int *array_v, int *array_j, int *array_coef, int array_size, int rhs);

/* Functions for internal use only! */

ECOVECptr ECOallocvector(int *status, int size);
void ECOfreevector(ECOVECptr *vec);
int ECOvsize(ECOVECptr vec);
void ECOvmaxsize(ECOVECptr vec);
void ECOvredim(ECOVECptr vec, int newsize);
int ECOvget(ECOVECptr vec, int i);
void ECOvset(ECOVECptr vec, int i, int val);
void ECOvincr(ECOVECptr vec, int i);
void ECOvdecr(ECOVECptr vec, int i);
void ECOvcopy(ECOVECptr vecsrc, ECOVECptr vecdest);
void ECOvfill(ECOVECptr vec, int val);
void ECOvflip(ECOVECptr vec);
void ECOshowvector(ECOVECptr vec);
int ECOsort(ECOVECptr index, ECOVECptr value, ECOVECptr ordering, int ordertype);
ECOMTXptr ECOallocmatrix(int *status, int size1, int size2);
void ECOfreematrix(ECOMTXptr *mtx);
int ECOmsize1(ECOMTXptr mtx);
int ECOmsize2(ECOMTXptr mtx);
int ECOmget(ECOMTXptr mtx, int i, int j);
void ECOmset(ECOMTXptr mtx, int i, int j, int val);
void ECOmfill(ECOMTXptr mtx, int val);
int ECOordersymmetry(struct ecointernal *ei);
int ECOlowerbound(struct ecointernal *ei);
int ECOalternatebound(struct ecointernal *ei, int *min_color);
int ECOgreedyclique(struct ecointernal *ei, int *max_clique);
int ECOmaxindepneighborhood(struct ecointernal *ei);
int ECOgetclique(struct ecointernal *ei, ECOVECptr ordering, ECOVECptr forbidden,
                 ECOVECptr clique_set, int flag);
int ECOgetindepset(struct ecointernal *ei, ECOVECptr ordering, ECOVECptr forbidden,
                   ECOVECptr indep_set);
int ECOgreedyindepset(struct ecointernal *ei, ECOVECptr list_set,
                      ECOVECptr indep_set, int *indep_size);
int ECOgreedypartition(struct ecointernal *ei, ECOVECptr list_set, int *partition_size);
int ECOinitcol(struct ecointernal *ei);
int ECOnaivecoloring(struct ecointernal *ei, ECOVECptr coloring);
int ECOfurmanczyk(struct ecointernal *ei, ECOVECptr ordering, ECOVECptr coloring, int colors_used);
int ECOprobgen(struct ecointernal *ei);
int ECOrvalue(struct ecointernal *ei, int v, int j, int min_color);
int ECOfloor(int n, int j);
int ECOceil(int n, int j);
int ECOtkj(int k, int j, int n);
double ECOxval(int v, int j, struct ecointernal *ei);
double ECOwval(int j, struct ecointernal *ei);
int ECOevalx(struct ecointernal *ei, ECOVECptr coloring);
int ECOgetsolution(struct ecointernal *ei);
int ECOinitdegrees(struct ecointernal *ei);
int ECOgetdegree(struct ecointernal *ei, int v);
int ECOgetneighbor(struct ecointernal *ei, int v, int i);
void ECOsetinternal(struct ecointernal *ei);
void ECOsetinternal_onlycallback(struct ecointernal *ei);
void ECOfreemem(struct ecointernal *ei);
void ECOfreemem_onlycallback(struct ecointernal *ei);
int ECOordervertices(struct ecointernal *ei);
int ECOcallbackinit(struct ecointernal *ei);
int ECOincumcallback(CPXCENVptr cenv, void *cbdata, int wherefrom,
                     struct ecointernal *ei, double *x, double objval);
int ECOnodecallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei);
int ECOcutcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                   int *useraction_p);
int ECOinitcutcallback(struct ecointernal *ei);
int ECOcutordering(struct ecointernal *ei, int j);
void ECOshowcutlog(struct ecointernal *ei);
int ECOinitcliquecallback(struct ecointernal *ei);
int ECOcliqueseparation(struct ecointernal *ei, int j);
int ECOaddcliquecut(struct ecointernal *ei, ECOVECptr clique_set, int j, int *cutadded);
int ECObuklval(struct ecointernal *ei, int u, int k, int l);
int ECOaddujkscut(struct ecointernal *ei, int u, int j, int k, ECOVECptr clique_set, int *cutadded);
int ECOinitalpha2callback(struct ecointernal *ei);
int ECOalpha2separation(struct ecointernal *ei, int j);
int ECOaddalpha2cut(struct ecointernal *ei, ECOVECptr alpha2_set,
                    ECOVECptr alpha2clique_set, int j, int *cutadded);
int ECOblockseparation(struct ecointernal *ei);
int ECOextraseparation(struct ecointernal *ei);
int ECOneighborhoodseparation(struct ecointernal *ei);
int ECOoutsideneighborseparation(struct ecointernal *ei);
int ECOinitcscallback(struct ecointernal *ei);
int ECOcsseparation(struct ecointernal *ei);
int ECOaddcscut(struct ecointernal *ei, ECOVECptr c_set, int *cutadded);
void ECObegincut(struct ecocut *cu, char sense, int rhs);
void ECOaddxcut(struct ecocut *cu, int v, int j, int coef);
void ECOaddwcut(struct ecocut *cu, int j, int coef);
int ECOendcut(struct ecocut *cu, int *cutadded, double threshold, int islocal);
int ECOinitheurcallback(struct ecointernal *ei);
int ECOheurcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                              double *objval_p, double *x, int *useraction_p);
int ECOprigreedy(struct ecointernal *ei, int current_k, int *fail);
int ECOpritabu(struct ecointernal *ei, int *fail);
int ECOmakefeasible(struct ecointernal *ei);
void ECOgivex(struct ecointernal *ei, double *x);
int ECOenum(struct ecointernal *ei);
int ECOenumbranch(struct ecointernal *ei, int lb);
int ECOcheckineqenum(struct ecointernal *ei, double start_t, double *pass_t);
int ECOtabuinit(struct ecointernal *ei);
int ECOtabueqcol(struct ecointernal *ei, double maxtim_t);
void ECOgeninitialsol(struct ecointernal *ei, ECOVECptr coloring, int colors_used, int current_k);
int ECOtabusearch(struct ecointernal *ei, int *iter, int maxiter, double start_t,
                  double maxtim_t, int current_k);
int ECOinitbranchcallback(struct ecointernal *ei);
int ECOenumcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                               int *useraction_p);
int ECObranchcallback(CPXCENVptr cenv, void *cbdata, int wherefrom, struct ecointernal *ei,
                               int *useraction_p);
#ifdef ECOPARALLEL
void ECOthreadinit(struct ecointernal *ei);
int ECOcreateeis(struct ecointernal *ei);
struct ecointernal *ECOgetei(struct ecointernal *ei);
void ECOfreememeis(struct ecointernal *ei);
int ECOparalleltabueqcol(struct ecointernal *ei, double maxtim_t);
#endif

/* for debugging */
#ifdef ECODEBUG
/*  #define ECODEBUGINIT */
/*  #define ECODEBUGPRIMAL */
/*  #define ECODEBUGCUTS */
/*  #define ECODEBUGBRANCH */
/*  #define ECODEBUGTABU */
  #define ECODEBUGENUM
/* #define ECODEF_WRITEROOTFILE */
/* #define ECODEF_WRITENODEFILE */
#endif

#ifdef __cplusplus
}
#endif

#endif /* __ECOPT_H */
