
/***
 * ecopt - Equitable Coloring Optimizer
 * Made in 2008 by Daniel E. Severin
 *
 * primal2.c - primal heuristic routines
 ***/

#include <ecopt.h>
#include <stdio.h>

/* TODO: polish this piece of code :P */

int ECOprigreedy(struct ecointernal *ei, int current_k, int *fail)
{
  ECOVECptr coloring, starting_coloring, size_classes, filled_color_set;
  ECOVECptr available_colors, neighborhood_colors, available_vertices;
  ECOVECptr available_qvert;

  int vertices, max_color, v, vv, j, k, degree, d, lb, p, r, size, l, flag;
  int count_filled_colors, count_available_colors, fixed_vertices, best_j, qvert;
  int count_neighborhood_colors, best_neighborhood, best_v, best_size, best_qvert;
  double xvj, best_val;

  if (!(ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_SCANXVJ) && !(ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_SCANCOLOR)) {
    /* if both heuristics are not defined, exit */
    *fail = 1;
    return ECOERR_OK;
  }

  /* set variables */
  vertices = ei->grp->vertices;
  max_color = ei->max_color;
  coloring = ei->primal_coloring;
  ECOvredim(coloring, vertices);
  starting_coloring = ei->primal_vector1;
  size_classes = ei->primal_vector2;
  filled_color_set = ei->primal_vector3;
  available_colors = ei->primal_vector4;
  neighborhood_colors = ei->primal_vector5;
  available_vertices = ei->primal_vector6;
  available_qvert = available_colors;
  lb = (int)(ei->LB_current + 1.0 - ECO_WJTOL);
  *fail = 1;

  /* check if optimality is reached */
  if (lb >= current_k) return ECOERR_OK;

  /* obtain a partial coloring from variables x*_vj = 1 */
  fixed_vertices = 0;
  ECOvredim(starting_coloring, vertices);
  ECOvfill(starting_coloring, 0);
  for (v = 1; v <= vertices; v++) {
    for (j = 1; j <= current_k - 1; j++) {
      if (ECOxval(v, j, ei) > 1.0 - ECO_EPSILON) {
	    ECOvset(starting_coloring, v, j);
	    fixed_vertices++;
		if (j > lb) lb = j;
	    break; /* <-- in order to continue with the next vertex */
      }
    }
  }

#ifdef ECODEBUGPRIMAL
  printf("Partial coloring (fixed vertices = %d):", fixed_vertices);
  ECOshowvector(starting_coloring);
#endif
/* if you want to limit the minimum number of colored vertices:
  if (fixed_vertices < (vertices * ECO_THRESHOLD_VNUM)/100) return ECOERR_OK; */

  /* main loop */
  for (k = lb; k <= current_k - 1; k++) {

    /********************************************* COMIENZO HEURISTICA 1 *********************************************************/
if (ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_SCANXVJ) {

	p = ECOfloor(vertices, k);
    r = vertices % k;
    ECOvcopy(starting_coloring, coloring);
    ECOvredim(size_classes, k);
    ECOvfill(size_classes, 0);
    ECOvredim(filled_color_set, k);
    ECOvfill(filled_color_set, 0);
    count_filled_colors = 0;

    /* count number of elements of each class (we scan with SL ordering) */
    for (vv = 1; vv <= vertices; vv++) {
      v = ECOvget(ei->sl_ordering, vv);
      j = ECOvget(coloring, v);
      if (j > 0) {
	if (ECOvget(filled_color_set, j) == 1) {
	  /* if class j is full, un-color that vertex */
	  ECOvset(coloring, v, 0);
	}
        else {
	  /* otherwise, increment its size */
          size = ECOvget(size_classes, j);
	  size++;
	  ECOvset(size_classes, j, size);
          if (count_filled_colors >= r) {
	    if (size == p) {
	      ECOvset(filled_color_set, j, 1);
	      count_filled_colors++;
	    }
	  }
	  else {
	    if (size == p + 1) {
	      ECOvset(filled_color_set, j, 1);
	      count_filled_colors++;
	      if (count_filled_colors == r) {
		/* from now on, every class with p elements are also full */
		for (l = 1; l <= k; l++) {
		  if (ECOvget(size_classes, l) == p) {
	            ECOvset(filled_color_set, l, 1);
	            count_filled_colors++;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
#ifdef ECODEBUGPRIMAL
    printf("STARTING COLORING:\n");
    for (j = 1; j <= k; j++) {
      if (ECOvget(filled_color_set, j) == 1) printf("Class %d is full, and has:", j);
      else printf("Class %d has:", j);
      for (v = 1; v <= vertices; v++) if (ECOvget(coloring, v) == j) printf(" %d", v);
      printf("\n");
    }
    printf("The number of filled classes is %d.\n", count_filled_colors);
#endif

    /* scan remaining vertices */
    ECOvredim(available_colors, k);
    ECOvredim(neighborhood_colors, k);
    do {
      best_v = 0;
      best_j = 0;
      best_neighborhood = -1;
      best_val = -1.0;
      for (v = 1; v <= vertices; v++) {
        if (ECOvget(coloring, v) > 0) continue;

        /* available_colors = { 1,...,k } - filled_color_set */
	count_available_colors = 0;
	for (j = 1; j <= k; j++) {
	  if (ECOvget(filled_color_set, j) == 1) ECOvset(available_colors, j, 0);
	  else {
            ECOvset(available_colors, j, 1);
	    count_available_colors++;
	  }
	}
	/* we also consider "unavailable" to the colors previously used by the neighbors of v
         * also, we fill the set of colors used in the neighborhood */
        ECOvfill(neighborhood_colors, 0);
	degree = ECOgetdegree(ei, v);
	for (d = 1; d <= degree; d++) {
	  j = ECOvget(coloring, ECOgetneighbor(ei, v, d));
	  if (j > 0) {
            ECOvset(neighborhood_colors, j, 1);
	    if (ECOvget(available_colors, j) == 1) {
	      ECOvset(available_colors, j, 0);
	      count_available_colors--;
              /* is there available colors yet? if not, skip to the next "k" */
	      if (count_available_colors == 0) goto next_iteration;
	    }
	  }
	}

        count_neighborhood_colors = 0;
        for (j = 1; j <= k; j++) if (ECOvget(neighborhood_colors, j) == 1) count_neighborhood_colors++;
	for (j = 1; j <= k; j++) {
	  if (ECOvget(available_colors, j) == 1) {
            /* we select the vertex and color according to the greatest value of x_vj */
	    xvj = ECOxval(v, j, ei);
#ifdef ECODEBUGPRIMAL
            printf("Vertex: %d  -  Color: %d  -  Value: %.2f  -  Neighborhood: %d  -  Available colors:", v, j, xvj, count_neighborhood_colors);
            for (d = 1; d <= k; d++) if (ECOvget(available_colors, d) == 1) printf(" %d", d);
            printf("\n");
#endif
	    if (xvj > best_val + ECO_EPSILON) {
              best_v = v;
	      best_j = j;
              best_neighborhood = count_neighborhood_colors;
	      best_val = xvj;
	    }
	    else {
	      /* in case of tie, we use two criteria: 1) select the vertex with the most colorful neighborhood
                                                      2) select the color corresponding to the smaller class */
	      if (xvj > best_val - ECO_EPSILON) {
                if (v == best_v) {
		  if (ECOvget(size_classes, j) < ECOvget(size_classes, best_j)) best_j = j;
                  /* (values best_v, best_val and best_neighborhood do not need to be updated) */
                }
                else {
                  if (count_neighborhood_colors > best_neighborhood) {
                    best_v = v;
                    best_j = j;
                    best_neighborhood = count_neighborhood_colors;
                    /* (value best_val does not need to be updated) */
                  }
                  else {
                    if (count_neighborhood_colors == best_neighborhood && ECOvget(size_classes, j) < ECOvget(size_classes, best_j)) {
                      best_v = v;
                      best_j = j;
                      /* (values best_val and best_neighborhood do not need to be updated) */
                    }
                  }
                }
	      }
	    }
	  }
	}
      }

#ifdef ECODEBUG
      if (best_v == -1) ECOabort(ECOERR_INTERNAL);
#endif
      /* at this point, "best_j" is the chosen color for v */
      ECOvset(coloring, best_v, best_j);
      size = ECOvget(size_classes, best_j);
      size++;
      ECOvset(size_classes, best_j, size);
      if (count_filled_colors >= r) {
        if (size == p) {
          ECOvset(filled_color_set, best_j, 1);
	  count_filled_colors++;
	}
      }
      else {
        if (size == p + 1) {
	  ECOvset(filled_color_set, best_j, 1);
	  count_filled_colors++;
          if (count_filled_colors == r) {
	    /* from now on, every class with p elements are also full */
	    for (l = 1; l <= k; l++) {
	      if (ECOvget(size_classes, l) == p) {
	        ECOvset(filled_color_set, l, 1);
	        count_filled_colors++;
	      }
	    }
	  }
	}
      }
#ifdef ECODEBUGPRIMAL
      printf("Vertex chosen: %d  -  Color chosen: %d\n", best_v, best_j);
      for (j = 1; j <= k; j++) {
        if (ECOvget(filled_color_set, j) == 1) printf("Class %d is full, and has:", j);
        else printf("Class %d has:", j);
        for (d = 1; d <= vertices; d++) if (ECOvget(coloring, d) == j) printf(" %d", d);
        printf("\n");
      }
      printf("The number of filled classes is %d.\n", count_filled_colors);
#endif
    } while (count_filled_colors < k);

#ifdef ECODEBUG
    for (v = 1; v <= vertices; v++) {
      if (ECOvget(coloring, v) == 0) {
        printf("Color of vertex %d not defined.\n", v);
        ECOabort(ECOERR_INTERNAL);
      }
    }
#endif
    /* at this point, we have a k-eqcol :) */
    ei->primal_colors_used = k;
    *fail = 0;
    CPXmsg(ei->env->cpxlog, "Primal heuristic 1 found a %d-eqcol.\n", k);
    CPXmsg(ei->env->cpxlog, "Elapsed time = %.2f sec.\n", ECOclock() - ei->start_t);
    return ECOERR_OK;

next_iteration:;
}

    /********************************************* COMIENZO HEURISTICA 2 *********************************************************/
if (ei->env->par[ECO_PARAM_PRIHEUR] & ECO_PRIHEUR_BIT_SCANCOLOR) {

	p = ECOfloor(vertices, k);
    r = vertices % k;
    ECOvcopy(starting_coloring, coloring);
    ECOvredim(size_classes, k);
    ECOvfill(size_classes, 0);
    ECOvredim(filled_color_set, k);
    ECOvfill(filled_color_set, 0);
    count_filled_colors = 0;

    /* count number of elements of each class (we scan with SL ordering) */
    for (vv = 1; vv <= vertices; vv++) {
      v = ECOvget(ei->sl_ordering, vv);
      j = ECOvget(coloring, v);
      if (j > 0) {
	if (ECOvget(filled_color_set, j) == 1) {
	  /* if class j is full, un-color that vertex */
	  ECOvset(coloring, v, 0);
	}
        else {
	  /* otherwise, increment its size */
          size = ECOvget(size_classes, j);
	  size++;
	  ECOvset(size_classes, j, size);
          if (count_filled_colors >= r) {
	    if (size == p) {
	      ECOvset(filled_color_set, j, 1);
	      count_filled_colors++;
	    }
	  }
	  else {
	    if (size == p + 1) {
	      ECOvset(filled_color_set, j, 1);
	      count_filled_colors++;
	      if (count_filled_colors == r) {
		/* from now on, every class with p elements are also full */
		for (l = 1; l <= k; l++) {
		  if (ECOvget(size_classes, l) == p) {
	            ECOvset(filled_color_set, l, 1);
	            count_filled_colors++;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
#ifdef ECODEBUGPRIMAL
    printf("STARTING COLORING:\n");
    for (j = 1; j <= k; j++) {
      if (ECOvget(filled_color_set, j) == 1) printf("Class %d is full, and has:", j);
      else printf("Class %d has:", j);
      for (v = 1; v <= vertices; v++) if (ECOvget(coloring, v) == j) printf(" %d", v);
      printf("\n");
    }
    printf("The number of filled classes is %d.\n", count_filled_colors);
#endif

    /* scan remaining vertices */
    ECOvredim(neighborhood_colors, k);
    ECOvredim(available_vertices, vertices);
    ECOvredim(available_qvert, vertices);
    do {
      /* first, we obtain the class color with less quantity of fractional x^*_vj.
       * in case of tie, we select the class with less quantity of vertices that can be colored with that class */
      best_j = 0;
      best_size = vertices + 1;
      best_qvert = vertices + 1;
      for (j = 1; j <= k; j++) {
        if (ECOvget(filled_color_set, j) == 0) {
          size = 0;
          for (v = 1; v <= vertices; v++) {
            xvj = ECOxval(v, j, ei);
            if (xvj > ECO_EPSILON && xvj < 1.0 - ECO_EPSILON) size++;
          }
#ifdef ECODEBUGPRIMAL
          printf("Class: %d  -  FracVars: %d\n", j, size);
#endif
          if (size <= best_size) {
            /* compute how many vertices can be colored with class j */
            qvert = 0;
            ECOvfill(available_qvert, 0);
            for (v = 1; v <= vertices; v++) {
              if (ECOvget(coloring, v) > 0) continue;
              flag = 0;
              degree = ECOgetdegree(ei, v);
              for (d = 1; d <= degree; d++) {
                if (ECOvget(coloring, ECOgetneighbor(ei, v, d)) == j) {
                  flag = 1;
                  break;
                }
              }
              if (flag == 0) {
                ECOvset(available_qvert, v, 1);
                qvert++;
              }
            }
            /* is there any available vertex for class j?, if not skip to the next "k" */
            if (qvert == 0) goto next_iteration2;
#ifdef ECODEBUGPRIMAL
            printf(" is candidate with available vertices: %d\n", qvert);
#endif
          }
          if (size < best_size) {
            best_j = j;
            best_size = size;
            ECOvcopy(available_qvert, available_vertices);
            best_qvert = qvert;
          }
          else {
            if (size == best_size && qvert < best_qvert) {
              best_j = j;
              ECOvcopy(available_qvert, available_vertices);
              best_qvert = qvert;
              /* (value best_size does not need to be updated) */
            }
          }
        }
      }
#ifdef ECODEBUG
      if (best_j == 0) ECOabort(ECOERR_INTERNAL);
#endif

      best_v = 0;
      best_neighborhood = -1;
      best_val = -1.0;
      for (v = 1; v <= vertices; v++) {
        if (ECOvget(available_vertices, v) == 0) continue;
        /* we select the vertex and color according to the greatest value of x_vj ... */
        xvj = ECOxval(v, best_j, ei);
#ifdef ECODEBUGPRIMAL
        printf("Vertex: %d  -  Color: %d  -  Value: %.2f\n", v, best_j, xvj);
#endif
        if (xvj > best_val - ECO_EPSILON) {
          /* ... but first, we fill the set of colors used in the neighborhood of a candidate vertex */
          ECOvfill(neighborhood_colors, 0);
	  degree = ECOgetdegree(ei, v);
	  for (d = 1; d <= degree; d++) {
	    j = ECOvget(coloring, ECOgetneighbor(ei, v, d));
	    if (j > 0) {
              ECOvset(neighborhood_colors, j, 1);
#ifdef ECODEBUG
              if (j == best_j) ECOabort(ECOERR_INTERNAL); 
#endif
	    }
	  }
          count_neighborhood_colors = 0;
          for (j = 1; j <= k; j++) if (ECOvget(neighborhood_colors, j) == 1) count_neighborhood_colors++;
#ifdef ECODEBUGPRIMAL
          printf(" is candidate with neighborhood: %d\n", count_neighborhood_colors);
#endif
        }
        if (xvj > best_val + ECO_EPSILON) {
          best_v = v;
          best_neighborhood = count_neighborhood_colors;
	  best_val = xvj;
	}
	else {
	  /* in case of tie, we select the vertex with the most colorful neighborhood */
	  if (xvj > best_val - ECO_EPSILON && count_neighborhood_colors > best_neighborhood) {
            best_v = v;
            best_neighborhood = count_neighborhood_colors;
            /* (value best_val does not need to be updated) */
          }
        }
      }

#ifdef ECODEBUG
      if (best_v == 0) ECOabort(ECOERR_INTERNAL);
#endif

      /* at this point, best_v is the chosen vertex for the color best_j */
      ECOvset(coloring, best_v, best_j);
      size = ECOvget(size_classes, best_j);
      size++;
      ECOvset(size_classes, best_j, size);
      if (count_filled_colors >= r) {
        if (size == p) {
	  ECOvset(filled_color_set, best_j, 1);
	  count_filled_colors++;
	}
      }
      else {
	if (size == p + 1) {
	  ECOvset(filled_color_set, best_j, 1);
	  count_filled_colors++;
          if (count_filled_colors == r) {
	    /* from now on, every class with p elements are also full */
	    for (l = 1; l <= k; l++) {
	      if (ECOvget(size_classes, l) == p) {
	        ECOvset(filled_color_set, l, 1);
	        count_filled_colors++;
	      }
	    }
	  }
	}
      }
#ifdef ECODEBUGPRIMAL
      printf("Vertex chosen: %d  -  Color chosen: %d\n", best_v, best_j);
      for (j = 1; j <= k; j++) {
        if (ECOvget(filled_color_set, j) == 1) printf("Class %d is full, and has:", j);
        else printf("Class %d has:", j);
        for (d = 1; d <= vertices; d++) if (ECOvget(coloring, d) == j) printf(" %d", d);
        printf("\n");
      }
      printf("The number of filled classes is %d.\n", count_filled_colors);
#endif
    } while (count_filled_colors < k);

#ifdef ECODEBUG
    for (v = 1; v <= vertices; v++) {
      if (ECOvget(coloring, v) == 0) {
        printf("Color of vertex %d not defined.\n", v);
        ECOabort(ECOERR_INTERNAL);
      }
    }
#endif
    /* at this point, we have a k-eqcol :) */
    ei->primal_colors_used = k;
    *fail = 0;
    CPXmsg(ei->env->cpxlog, "Primal heuristic 2 found a %d-eqcol.\n", k);
    CPXmsg(ei->env->cpxlog, "Elapsed time = %.2f sec.\n", ECOclock() - ei->start_t);
    return ECOERR_OK;
next_iteration2:;
}
  }

  /* no coloring was found :( */
  return ECOERR_OK;
}

int ECOpritabu(struct ecointernal *ei, int *fail)
{
  int iter, status, current_k;
  double start_t;

  /* we assume a primal coloring exists */
  current_k = ei->primal_colors_used - 1;
  *fail = 1;
  start_t = ECOclock();

  /* obtain initial solution */
  ECOgeninitialsol(ei, ei->primal_coloring, ei->primal_colors_used, current_k);
  ECOvcopy(ei->tabu_curcoloring, ei->tabu_bestcoloring);

  /* make the tabu search */
  CPXmsg(ei->env->cpxlog, "We are going to find a %d-eqcol.\n", current_k);
  status = ECOtabusearch(ei, &iter, ECO_MAXITERPRI, start_t, ECO_MAXTIMEPRITABU, current_k);
  ei->totaliter += iter;

  if (status == ECOTABU_OK) {
    ECOvcopy(ei->tabu_bestcoloring, ei->primal_coloring);
    ei->primal_colors_used = current_k;
    *fail = 0;
    CPXmsg(ei->env->cpxlog, "Tabueqcol found a %d-eqcol in %d iterations.\n", current_k, iter);
  }

  return ECOERR_OK;
}
