#!/usr/bin/perl -w
use strict;

# Undirected random graph generator.
#
# (c) 2008 Daniel E. Severin.

print "grandom - Undirected random graph generator\n\n";

sub manual {
	print "Usage: ./grandom.pl file.col vertices prob\n";
	print "vertices may be a value between 0 and 5000.\n";
	print "prob may be a value between 0 and 100.\n";
	exit 1;
}

if ($#ARGV < 2) { manual; }

my $file_col = $ARGV[0];
my $vertices = $ARGV[1];
my $prob = $ARGV[2];

my $minvertices = 3;
my $maxvertices = 5000;

if ($vertices < $minvertices || $vertices > $maxvertices) { manual; }

if ($prob < 0 || $prob > 100) { manual; }

open OUT, '>', 'edges.temp' or die "Error creating file: $!";
my $v1;
my $v2;
my $edges = 0;
for ($v1 = 1; $v1 < $vertices; $v1++) {
	for ($v2 = $v1 + 1; $v2 <= $vertices; $v2++) {
		my $aleat = int(rand(100));
		if ($aleat < $prob) {
			print OUT "e $v1 $v2\n";
			$edges++;
			print OUT "e $v2 $v1\n";
			$edges++;
		}
	}
}
close OUT;

open IN, '<', 'edges.temp' or die "Error reading file: $!";
print "Output file: $file_col\n";
open OUT, '>', $file_col or die "Error creating file: $!";

print OUT "c File generated with grandom script.\n";
print OUT "c Made by Daniel Severin.\n";
print OUT "p edges $vertices $edges\n";
my $e;
for ($e = 1; $e <= $edges; $e++) {
	$_ = <IN>;
	chomp;
	print OUT "$_\n";
}

close OUT;
close IN;

unlink 'edges.temp';

print "|V| = $vertices, |E| = " . $edges / 2 . ".\n";
print "Ready!\n";
exit 0;
